@@ -1,7 +1,7 @@
 # -*- perl -*-
 
 #
-# $Id: HistEntry.pm,v 1.10 1998/05/22 13:27:09 eserte Exp $
+# $Id: HistEntry.pm,v 1.14 1999/03/17 18:51:37 eserte Exp $
 # Author: Slaven Rezic
 #
 # Copyright  1997 Slaven Rezic. All rights reserved.
@@ -17,51 +17,71 @@ require Tk;
 use strict;
 use vars qw($VERSION);
 
-$VERSION = '0.22';
+$VERSION = '0.30';
 
 sub addBind {
     my $w = shift;
-    $w->bind('<Up>'        => sub { $w->historyUp });
-    $w->bind('<Control-p>' => sub { $w->historyUp });
-    $w->bind('<Down>'      => sub { $w->historyDown });
-    $w->bind('<Control-n>' => sub { $w->historyDown });
 
-    $w->bind('<Meta-less>'    => sub { $w->historyBegin });
-    $w->bind('<Alt-less>'     => sub { $w->historyBegin });
-    $w->bind('<Meta-greater>' => sub { $w->historyEnd });
-    $w->bind('<Alt-greater>'  => sub { $w->historyEnd });
+    $w->_entry->bind('<Up>'        => sub { $w->historyUp });
+    $w->_entry->bind('<Control-p>' => sub { $w->historyUp });
+    $w->_entry->bind('<Down>'      => sub { $w->historyDown });
+    $w->_entry->bind('<Control-n>' => sub { $w->historyDown });
 
-    $w->bind('<Control-r>' => sub { $w->searchBack });
-    $w->bind('<Control-s>' => sub { $w->searchForw });
+    $w->_entry->bind('<Meta-less>'    => sub { $w->historyBegin });
+    $w->_entry->bind('<Alt-less>'     => sub { $w->historyBegin });
+    $w->_entry->bind('<Meta-greater>' => sub { $w->historyEnd });
+    $w->_entry->bind('<Alt-greater>'  => sub { $w->historyEnd });
 
-    $w->bind('<Return>' => sub { 
-		 if ($w->cget(-command)) {
+    $w->_entry->bind('<Control-r>' => sub { $w->searchBack });
+    $w->_entry->bind('<Control-s>' => sub { $w->searchForw });
+
+    $w->_entry->bind('<Return>' => sub { 
+		 if ($w->cget(-command) || $w->cget(-auto)) {
 		     $w->invoke;
 		 }
 	     });
+
+    $w->_entry->bind('<Any-KeyPress>', sub { my $e = $_[0]->XEvent;
+					     $w->KeyPress($e->K) });
 }
 
 sub _isdup {
     my($w, $string) = @_;
-    foreach (@{$w->{'history'}}) {
+    foreach (@{ $w->privateData->{'history'} }) {
 	return 1 if $_ eq $string;
     }
     0;
 }
 
+sub _update {
+    my($w, $string) = @_;
+    $w->_entry->delete(0, 'end');
+    $w->_entry->insert('end', $string);
+}
+
+sub _entry {
+    my $w = shift;
+    $w->Subwidget('entry') ? $w->Subwidget('entry') : $w;
+}
+
+sub _listbox {
+    my $w = shift;
+    $w->Subwidget('slistbox') ? $w->Subwidget('slistbox') : $w;
+}
+
 sub historyAdd {
     my($w, $string) = @_;
-    $string = $ {$w->cget(-textvariable)} if !defined $string;
+    $string = $w->_entry->get unless defined $string;
     return undef if !defined $string || $string eq '';
-    if ((!@{$w->{'history'}}
-	 || $string ne $w->{'history'}->[$#{$w->{'history'}}])
-	&& ($w->cget(-dup) || !$w->_isdup($string))) {
-	push(@{$w->{'history'}}, $string);
+    my $history_ref = $w->privateData->{'history'};
+    if ((!@{ $history_ref } or $string ne $history_ref->[$#$history_ref])
+	and ($w->cget(-dup) or !$w->_isdup($string))) {
+	push @{ $history_ref }, $string;
 	if (defined $w->cget(-limit) &&
-	    @{$w->{'history'}} > $w->cget(-limit)) {
-	    shift @{$w->{'history'}};
+	    @{ $history_ref } > $w->cget(-limit)) {
+	    shift @{ $history_ref };
 	}
-	$w->{'historyindex'} = $#{$w->{'history'}} + 1;
+	$w->privateData->{'historyindex'} = $#$history_ref + 1;
 	return $string;
     }
     undef;
@@ -71,14 +91,14 @@ sub historyAdd {
 
 sub historyUpdate {
     my $w = shift;
-    $ {$w->cget(-textvariable)} = $w->{'history'}->[$w->{'historyindex'}];
-    $w->icursor('end'); # suggestion by Jason Smith <smithj4@rpi.edu>
+    $w->_update($w->privateData->{'history'}->[$w->privateData->{'historyindex'}]);
+    $w->_entry->icursor('end'); # suggestion by Jason Smith <smithj4@rpi.edu>
 }
 
 sub historyUp {
     my $w = shift;
-    if ($w->{'historyindex'} > 0) {
-	$w->{'historyindex'}--;
+    if ($w->privateData->{'historyindex'} > 0) {
+        $w->privateData->{'historyindex'}--;
 	$w->historyUpdate;
     } else {
 	$w->_bell;
@@ -87,8 +107,8 @@ sub historyUp {
 
 sub historyDown {
     my $w = shift;
-    if ($w->{'historyindex'} <= $#{$w->{'history'}}) {
-	$w->{'historyindex'}++;
+    if ($w->privateData->{'historyindex'} <= $#{$w->privateData->{'history'}}) {
+	$w->privateData->{'historyindex'}++;
 	$w->historyUpdate;
     } else {
 	$w->_bell;
@@ -97,45 +117,53 @@ sub historyDown {
 
 sub historyBegin {
     my $w = shift;
-    $w->{'historyindex'} = 0;
+    $w->privateData->{'historyindex'} = 0;
     $w->historyUpdate;
 }
 
 sub historyEnd {
     my $w = shift;
-    $w->{'historyindex'} = $#{$w->{'history'}};
+    $w->privateData->{'historyindex'} = $#{$w->privateData->{'history'}};
     $w->historyUpdate;
 }
 
 sub historySet {
     my($w, $index) = @_;
     my $i;
-    for($i = $#{$w->{'history'}}; $i >= 0; $i--) {
-	if ($index eq $w->{'history'}->[$i]) {
-	    $w->{'historyindex'} = $i;
+    my $history_ref = $w->privateData->{'history'};
+    for($i = $#{ $history_ref }; $i >= 0; $i--) {
+	if ($index eq $history_ref->[$i]) {
+	    $w->privateData->{'historyindex'} = $i;
 	    last;
 	}
     }
 }
 
+sub historyReset {
+    my $w = shift;
+    $w->privateData->{'history'} = [];
+    $w->privateData->{'historyindex'} = 0;
+}
+
 sub history {
     my($w, $history) = @_;
     if (defined $history) {
-	@{$w->{'history'}}   = @$history;
-	$w->{'historyindex'} = $#{$w->{'history'}} + 1;
+	$w->privateData->{'history'} = [ @$history ];
+	$w->privateData->{'historyindex'} =
+	  $#{$w->privateData->{'history'}} + 1;
     }
-    @{$w->{'history'}};
+    @{ $w->privateData->{'history'} };
 }
 
 sub searchBack {
     my $w = shift;
-    my $i = $w->{'historyindex'}-1;
+    my $i = $w->privateData->{'historyindex'}-1;
     while ($i >= 0) {
-	my $search = $ {$w->cget(-textvariable)};
-        if ($search eq substr($w->{'history'}->[$i], 0, length($search))) {
-	    $w->{'historyindex'} = $i;
-	    $ {$w->cget(-textvariable)} 
-                = $w->{'history'}->[$w->{'historyindex'}];
+	my $search = $w->_entry->get;
+        if ($search eq substr($w->privateData->{'history'}->[$i], 0,
+			      length($search))) {
+	    $w->privateData->{'historyindex'} = $i;
+	    $w->_update($w->privateData->{'history'}->[$w->privateData->{'historyindex'}]);
             return;
         }
         $i--;
@@ -145,13 +173,13 @@ sub searchBack {
 
 sub searchForw {
     my $w = shift;
-    my $i = $w->{'historyindex'}+1;
-    while ($i <= $#{$w->{'history'}}) {
-	my $search = $ {$w->cget(-textvariable)};
-        if ($search eq substr($w->{'history'}->[$i], 0, length($search))) {
-	    $w->{'historyindex'} = $i;
-	    $ {$w->cget(-textvariable)} 
-                = $w->{'history'}->[$w->{'historyindex'}];
+    my $i = $w->privateData->{'historyindex'}+1;
+    while ($i <= $#{$w->privateData->{'history'}}) {
+	my $search = $w->_entry->get;
+        if ($search eq substr($w->privateData->{'history'}->[$i], 0,
+			      length($search))) {
+	    $w->privateData->{'historyindex'} = $i;
+	    $w->_update($w->privateData->{'history'}->[$w->privateData->{'historyindex'}]);
             return;
         }
         $i++;
@@ -161,10 +189,10 @@ sub searchForw {
 
 sub invoke {
     my($w, $string) = @_;
-    $string = $ {$w->cget(-textvariable)} if !defined $string;
+    $string = $w->_entry->get if !defined $string;
     return unless defined $string;
     my $added = defined $w->historyAdd($string);
-    &{$w->cget(-command)}($w, $string, $added);
+    $w->Callback(-command => $w, $string, $added);
 }
 
 sub _bell {
@@ -173,106 +201,171 @@ sub _bell {
     $w->bell;
 }
 
+sub KeyPress {
+    my($w, $key) = @_;
+    my $e = $w->_entry;
+    my(@history) = reverse $w->history;
+    $w->{end} = $#history; # XXXXXXXX?
+    return if ($key =~ /^Shift|^Control|^Left|^Right/);
+    if ($key eq 'Tab') {
+	# Tab doesn't trigger FocusOut event so clear selection
+	$e->selection('clear');
+	return;
+    }
+  
+    return if (!$w->cget(-match));
+
+    $e->update;
+    my $cursor = $e->index('insert');
+
+    if ($key eq 'BackSpace') {
+	$w->{start} = 0;
+	$w->{end} = $#history;
+	return;
+    }
+  
+    my $text = $e->get;
+    ###Grab test from entry upto cursor
+    (my $typedtext = $text) =~ s/^(.{$cursor}).*/$1/;
+  
+    if ($cursor == 0 || $text eq '') {
+	###No text before cursor, reset list
+	$w->{start} = 0;
+	$w->{end} = $#history;
+	$e->delete(0, 'end');
+	$e->insert(0,'');
+    } else {
+	my $start = $w->{start};
+	my $end = $w->{end};
+	my ($newstart, $newend);
+
+	###Locate start of matching & end of matching
+	for (; $start <= $end; $start++) {
+	    if ($history[$start] =~ /^\Q$typedtext\E/) {
+		$newstart = $start if (!defined $newstart);
+		$newend = $start;
+	    } else {
+		last if (defined $newstart);
+	    }
+	}
+	
+	if (defined $newstart) {
+	    $e->selection('clear');
+	    $e->delete(0, 'end');
+	    $e->insert(0, $history[$newstart]);
+	    $e->selection('range',$cursor,'end');
+	    $e->icursor($cursor);
+	    $w->{start} = $newstart;
+	    $w->{end} = $newend;
+	} else {
+	    $w->{end} = -1;
+	} 
+    }
+}
+
 ######################################################################
 
 package Tk::HistEntry::Simple;
 require Tk::Entry;
 use vars qw(@ISA);
-@ISA = qw(Tk::Derived Tk::HistEntry Tk::Entry);
+@ISA = qw(Tk::Derived Tk::Entry Tk::HistEntry);
+#use base qw(Tk::Derived Tk::Entry Tk::HistEntry);
 Construct Tk::Widget 'SimpleHistEntry';
 
 sub Populate {
     my($w, $args) = @_;
 
-    $w->{'history'} = [];
-    $w->{'historyindex'} = 0;
-
-    $args->{'-textvariable'} = delete $args->{'-variable'} 
-        if defined $args->{'-variable'};
+    $w->historyReset;
 
     $w->SUPER::Populate($args);
 
+    $w->Advertise(entry => $w);
+
+    $w->{start} = 0;
+    $w->{end} = 0;
+
+    $w->addBind;
+
     $w->ConfigSpecs
       (-command => ['CALLBACK', 'command', 'Command', undef],
+       -auto    => ['PASSIVE',  'auto',    'Auto',    0],
        -dup     => ['PASSIVE',  'dup',     'Dup',     1],
        -bell    => ['PASSIVE',  'bell',    'Bell',    1],
        -limit   => ['PASSIVE',  'limit',   'Limit',   undef],
+       -match   => ['PASSIVE',  'match',   'Match',   0],
       );
 
     $w;
 }
 
-sub SetBindtags {
-    my($w) = @_;
-    $w->addBind;
-    $w->SUPER::SetBindtags;
-}
 
 ######################################################################
 package Tk::HistEntry::Browse;
 require Tk::BrowseEntry;
 use vars qw(@ISA);
-@ISA = qw(Tk::Derived Tk::HistEntry Tk::BrowseEntry);
+@ISA = qw(Tk::Derived Tk::BrowseEntry Tk::HistEntry);
+#use base qw(Tk::Derived Tk::BrowseEntry Tk::HistEntry);
 Construct Tk::Widget 'HistEntry';
 
 sub Populate {
     my($w, $args) = @_;
 
-    $w->{'history'} = [];
-    $w->{'historyindex'} = 0;
+    $w->historyReset;
 
-    $args->{'-variable'} = delete $args->{'-textvariable'} 
-        if defined $args->{'-textvariable'};
-
-#    $args->{'-browsecmd'} = sub { $w->historySet($_[1]) };
-
-    my $saveargs;
-    foreach (qw(-command -dup -bell -limit)) {
-	if (exists $args->{$_}) {
-	    $saveargs->{$_} = delete $args->{$_};
+    if ($Tk::VERSION >= 800) {
+	$w->SUPER::Populate($args);
+    } else {
+	my $saveargs;
+	foreach (qw(-auto -command -dup -bell -limit -match)) {
+	    if (exists $args->{$_}) {
+		$saveargs->{$_} = delete $args->{$_};
+	    }
+	}
+	$w->SUPER::Populate($args);
+	foreach (keys %$saveargs) {
+	    $args->{$_} = $saveargs->{$_};
 	}
     }
-    $w->SUPER::Populate($args);
-    foreach (keys %$saveargs) {
-	$args->{$_} = $saveargs->{$_};
-    }
+
+    $w->addBind;
+
+    $w->{start} = 0;
+    $w->{end} = 0;
 
     $w->ConfigSpecs
       (-command => ['CALLBACK', 'command', 'Command', undef],
+       -auto    => ['PASSIVE',  'auto',    'Auto',    0],
        -dup     => ['PASSIVE',  'dup',     'Dup',     0],
        -bell    => ['PASSIVE',  'bell',    'Bell',    1],
        -limit   => ['PASSIVE',  'limit',   'Limit',   undef],
+       -match   => ['PASSIVE',  'match',   'Match',   0],
       );
 
     $w;
 }
 
-sub SetBindtags {
-    my($w) = @_;
-    $w->addBind;
-    $w->SUPER::SetBindtags;
-}
-
 sub historyAdd {
     my($w, $string) = @_;
     if (defined($string = $w->SUPER::historyAdd($string))) {
 	$w->insert('end', $string);
+	# XXX Obeying -limit also for the array itself?
 	if (defined $w->cget(-limit) &&
-	    $w->Subwidget('slistbox')->size > $w->cget(-limit)) {
-	    $w->delete(0);
+	    $w->_listbox->size > $w->cget(-limit)) {
+	    $w->_entry->delete(0);
 	}
-	$w->Subwidget('slistbox')->see('end');
+	$w->_listbox->see('end');
 	return $string;
     }
     undef;
 }
+*addhistory = \&historyAdd;
 
 sub history {
     my($w, $history) = @_;
     if (defined $history) {
-	$w->Subwidget('slistbox')->delete(0, 'end');
-	$w->Subwidget('slistbox')->insert('end', @$history);
-	$w->Subwidget('slistbox')->see('end');
+	$w->_listbox->delete(0, 'end');
+	$w->_listbox->insert('end', @$history);
+	$w->_listbox->see('end');
     }
     $w->SUPER::history($history);
 }
@@ -349,6 +442,10 @@ or if a search was not successful. Defaults to true.
 
 Limits the number of history entries. Defaults to unlimited.
 
+=item B<-match>
+
+Turns auto-completion on.
+
 =back
 
 =head1 METHODS
@@ -432,6 +529,13 @@ executes the associated callback.
 
 Slaven Rezic <eserte@cs.tu-berlin.de>
 
+=head1 CREDITS
+
+Thanks for Jason Smith <smithj4@rpi.edu> and Benny Khoo
+<kkhoo1@penang.intel.com> for their suggestions. The auto-completion
+code is stolen from Tk::IntEntry by Dave Collins
+<Dave.Collins@tiuk.ti.com>.
+
 =head1 COPYRIGHT
 
 Copyright (c) 1997 Slaven Rezic. All rights reserved.
